from remotemanager.serialisation.serialyaml import serialyaml
from remotemanager.serialisation.serialjson import serialjson

__all__ = ['serialyaml', 'serialjson']

try:
    from remotemanager.serialisation.serialdill import serialdill  # noqa: F401
    __all__.append('serialdill')
except ImportError:
    pass

try:
    from remotemanager.serialisation.serialjsonpickle import serialjsonpickle  # noqa: F401, E501
    __all__.append('serialjsonpickle')
except ImportError:
    pass



import copy
import json
import os
import time

from remotemanager.storage.database import Database
from remotemanager.storage import SendableMixin, TrackedFile
from remotemanager.storage.sendablemixin import SERIALISED_STORAGE_KEY
from remotemanager.logging.utils import format_iterable
from remotemanager.utils.uuid import generate_uuid
from remotemanager.utils import object_from_uuid, _time_format
from remotemanager.logging import LoggingMixin
from remotemanager.logging.verbosity import Verbosity

from datetime import datetime

localwinerror = """Local runs on windows machines are not supported. 
Please use a URL which connects to a non-windows machine or consider using 
Docker to continue."""


class Runner(SendableMixin, LoggingMixin):
    """
    The Runner class stores any info pertaining to this specific run. E.g.
    Arguments, result, run status, files, etc.

    .. warning::
        Interacting with this object directly could cause unstable behaviour.
        It is best to allow Dataset to handle the runners. If you require a
        single run, you should create a Dataset and append just that one run.
    """

    _defaults = {'skip': True}

    _default_local_dir = 'temp_runner_local'
    _default_remote_dir = 'temp_runner_remote'

    _submit_status = {'initial': 'creating files and submitting...',
                      'final': 'command executed remotely'}

    _args_replaced_key = '~serialised_args~'

    _do_not_package = ['_serialiser', '_parent', '_database']

    def __init__(self,
                 arguments: dict,
                 dbfile: str,
                 parent,
                 self_id: str,
                 extra_files_send: list = None,
                 extra_files_recv: list = None,
                 verbose: int = None,
                 **kwargs):

        self._verbose = Verbosity(verbose)

        self._run_options = self._set_defaults(kwargs)

        self._extra_files = {'send': extra_files_send
                             if extra_files_send is not None else [],
                             'recv': extra_files_recv
                             if extra_files_recv is not None else []}

        if arguments is not None \
                and not isinstance(arguments, dict):
            raise ValueError(f'runner arguments ({type(arguments)}) '
                             'must be dict-type')

        # parent and id setting
        self._parent = parent
        self._parent_uuid = parent.uuid  # used for parent memory recovery

        self._id = self_id

        # check that we can properly serialise the args
        # this needs to be within the runner, so we can properly generate uuids
        self._args_replaced = False
        try:
            json.dumps(arguments)
            uuid_slug = copy.deepcopy(arguments) or {}
        except TypeError:
            file = f'{self.parent.argfile}-' \
                   f'{self.id}{self.serialiser.extension}'
            lpath = os.path.join(self.parent.local_dir, file)

            if not os.path.isdir(self.parent.local_dir):
                os.makedirs(self.parent.local_dir)

            content = self.parent.serialiser.dumps(arguments)
            with open(lpath, self.serialiser.write_mode) as o:
                o.write(content)

            # adding the file in here forces the kwargs to swap out
            # kwargs for a repo.load
            arguments = {file: Runner._args_replaced_key}
            self._args_replaced = True

            uuid_slug = {'uuid_base': generate_uuid(str(content))}

            extra_files_send.append(lpath)

        # uuid generation
        uuid_slug.update(**kwargs)
        self._runner_uuid = generate_uuid(format_iterable(uuid_slug))
        self._uuid = generate_uuid(self._runner_uuid + str(self.parent.uuid))

        self._logger.info(f'new runner (id {self.uuid}) created')

        self._args = arguments
        self._state_time = None
        self._last_submitted = -1
        self._state = None
        self._extension = 'yaml'

        self._dependency_info = {}

        self._dbfile = dbfile

        self._manual_run = False
        self._history = {}
        self.state = 'created'

        self._identifier = f'{self.parent.name}-' \
                           f'{self.parent.short_uuid}-{self.id}'

    def __hash__(self) -> hash:
        return hash(self.uuid)

    def __repr__(self) -> str:
        return self.identifier

    @property
    def verbose(self) -> Verbosity:
        return self._verbose

    @property
    def database(self) -> Database:
        """
        Access to the stored database object.
        Creates a connection if none exist.

        Returns:
            Database
        """
        if not hasattr(self, '_database'):
            self._database = Database(file=self._dbfile)
        return self._database

    @property
    def parent(self):
        if self.is_missing('_parent'):
            self._parent = object_from_uuid(self._parent_uuid, 'Dataset')
        return self._parent

    @property
    def serialiser(self):
        return self.parent.serialiser

    @staticmethod
    def _set_defaults(kwargs: dict = None) -> dict:
        """
        Sets default arguments as expected. If used as a staticmethod, returns
        the defaults
        """

        if kwargs is None:
            kwargs = {}

        for k, v in Runner._defaults.items():
            if k not in kwargs:
                kwargs[k] = v

        return kwargs

    @property
    def uuid(self) -> str:
        """
        The uuid of this runner
        """
        return self._uuid

    @property
    def short_uuid(self) -> str:
        """
        A short uuid for filenames
        """
        return self.uuid[:8]

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._id

    @property
    def identifier(self) -> str:
        return self._identifier

    def _format_filename(self,
                         identifier: str,
                         ftype: str,
                         ext: str) -> str:
        """
        Formats internal file names consistently.

        Args:
            identifier (str):
                dataset name and runner id as string
            ftype (str):
                file type. Jobscript, result file, etc.
            ext (str):
                file extension

        Returns:
            str: formatted filename
        """
        self._logger.info(f'creating filename for content {identifier}, '
                          f'filetype {ftype} and '
                          f'extension {ext}')
        return f'{identifier}-{ftype}{ext}'

    @property
    def runfile(self) -> TrackedFile:
        """
        Filename of the python runfile
        """
        base_name = self._format_filename(self.identifier, 'run', '.py')

        return TrackedFile(self.local_dir, self.remote_dir, base_name)

    @property
    def jobscript(self) -> TrackedFile:
        """
        Filename of the run script
        """
        base_name = self._format_filename(self.identifier, 'jobscript', '.sh')

        return TrackedFile(self.local_dir, self.remote_dir, base_name)

    @property
    def resultfile(self) -> TrackedFile:
        """
        Result file name
        """
        base_name = self._format_filename(self.identifier,
                                          'result',
                                          f'.{self.result_extension}')

        paths = [self.remote_dir]
        if self.run_dir != self.remote_dir and self.run_dir is not None:
            paths.append(self.run_dir)
        joined = '/'.join(paths)

        return TrackedFile(self.local_dir, joined, base_name)

    def _relative_remote_dir(self, file: str) -> str:
        """
        remote dir path relative to run dir
        """
        self._logger.info(f'getting relative path to remote dir for file '
                          f'{file} where remote={self.remote_dir} and run='
                          f'{self.run_dir}')
        if self.run_dir is None:
            self._logger.info('no run dir, returning file as rel path')
            return file
        return os.path.join('..', file)

    @property
    def result_extension(self) -> str:
        """
        Resultfile file format extension
        """
        return self._extension

    @result_extension.setter
    def result_extension(self, ext: str) -> None:
        """
        Sets the resultfile format extension

        .. warning::
            This does not change anything aside from the extension that
            the runner looks for when trying to find a result file. If you
            require a different serialisation, you should set the serialiser.
        """
        self._extension = ext.strip('.')

    @property
    def local_dir(self) -> str:
        """
        Local staging directory
        """
        return self._run_options.get('local_dir', Runner._default_local_dir)

    @local_dir.setter
    def local_dir(self,
                  path: str) -> None:
        """
        Sets the local_dir
        """
        self._run_options['local_dir'] = path

    @property
    def remote_dir(self) -> str:
        """
        Target directory on the remote for transports
        """
        if 'remote_dir' in self._run_options:
            return self._run_options['remote_dir']
        return self._run_options.get('run_dir', Runner._default_remote_dir)

    @remote_dir.setter
    def remote_dir(self,
                   path: str) -> None:
        """
        Sets the remote_dir
        """
        self._logger.debug(f'setting remote dir to {path}')
        self._run_options['remote_dir'] = path

    @property
    def run_dir(self) -> [str, None]:
        """
        Intended running directory. If not set, uses remote_dir

        .. note::
            If both remote_dir and run_dir are set, the files will be
            transferred to remote_dir, and then executed within run_dir
        """
        if 'run_dir' in self._run_options:
            abs_rundir = self._run_options['run_dir']
            common = os.path.commonpath((self.remote_dir, abs_rundir))

            run = os.path.split(abs_rundir.replace(common, ''))[1]
            return run
        return None

    @run_dir.setter
    def run_dir(self,
                dir: str) -> None:
        """
        Sets the run_dir
        """
        self._run_options['run_dir'] = dir

    @property
    def args(self) -> dict:
        """
        Arguments for the function
        """
        if self._args is None:
            return {}
        return self._args

    @property
    def extra_files(self) -> dict:
        """
        Returns the extra files set for this runner
        """
        return self._extra_files

    @property
    def result(self):
        """
        Result (If available)
        """
        if hasattr(self, '_result'):
            try:
                if SERIALISED_STORAGE_KEY in self._result:
                    self._result = self.parent.serialiser.loads(
                        self._result[1])
            except TypeError:
                pass

            return self._result
        return None

    @result.setter
    def result(self, result) -> None:
        """
        Creates and sets the result property, setting the state to "completed"

        Args:
            result:
                run result
        """
        self._result = result
        self.state = 'completed'

    def clear_result(self) -> None:
        """
        Removes any results, and sets the state to "result wiped"
        """
        try:
            del self._result
        except AttributeError:
            pass

        def remove_file(path):
            self._logger.info(f'attempting to clear result file {path}')
            try:
                os.remove(path)
                self._logger.info('Done')
            except FileNotFoundError:
                self._logger.info('file not found')

        remove_file(self.resultfile.local)
        remove_file(self.resultfile.remote)

        self.state = 'result wiped'

    @property
    def state(self) -> str:
        """
        Returns the most recent runner state
        """
        return self._state

    @state.setter
    def state(self,
              newstate: str) -> None:
        """
        Update the state and store within the runner history
        """

        state_time = datetime.now()

        self.insert_history(state_time, newstate)

        t = int(time.time())
        self._state_time = t

        if newstate == Runner._submit_status['final']:
            self._last_submitted = t
        self._state = newstate

    @property
    def last_updated(self) -> int:
        """
        Time that this runner state last changed
        """
        return self._state_time

    @property
    def last_submitted(self) -> int:
        """
        Time that this runnerwas last submitted
        """
        return self._last_submitted

    def format_time(self, t: datetime.time) -> str:
        """
        Format the datetime object into a dict key

        Args:
            t (datetime.time):
                time object to be formatted to string

        Returns:
            (str):
                formatted time
        """
        return t.strftime(_time_format)

    @property
    def history(self) -> dict:
        """
        State history of this runner
        """
        return self._history

    @property
    def status_list(self) -> list:
        """
        Returns a list of status updates
        """
        return list(self._history.values())

    def insert_history(self,
                       t: datetime,
                       newstate: str) -> None:
        """
        Insert a state into this runner's history

        Args:
            t (datetime.time):
                time this state change occurred
            newstate (str):
                status to update
        """
        if not isinstance(t, datetime):
            raise ValueError(f'time of type {type(t)} should be a datetime '
                             f'instance')

        base_timekey = self.format_time(t)
        idx = 0
        timekey = f'{base_timekey}/{idx}'
        while timekey in self._history:
            idx += 1

            timekey = f'{base_timekey}/{idx}'
            self._logger.info(f'timekey updated to {timekey}')

        self._logger.info(f'updating runner {self.short_uuid} state -> '
                          f'{newstate}')
        self._history[timekey] = newstate

    def run(self,
            dry_run: bool = False,
            **kwargs) -> None:
        """
        Perform a manual run

        .. warning::
            This method should be used sparingly, as it creates a Datset
            object within the function from the Database. This is a costly
            process and potentially unstable.

        Args:
            dry_run (bool):
                create files, but do not run
        """

        parent = self.unserialise(self.database.find(self.parent.uuid))

        if not self._assess_run(**self._run_options):
            return None

        if os.name == 'nt' and parent.url.is_local:
            raise RuntimeError(localwinerror)

        self._manual_run = True  # set internal flag for a manual run
        self._run_options.update(kwargs)
        self.state = Runner._submit_status['initial']
        self._write_runfile(parent)

        parent.transport.queue_for_push(self.runfile.name,
                                        self.local_dir,
                                        self.remote_dir)

        script = parent._script_sub(**self.run_args)

        self._write_script(parent.url.python,
                           script)

        parent.transport.queue_for_push(self.jobscript.name,
                                        self.jobscript.local_dir,
                                        self.jobscript.remote_dir)

        cmd = f'cd {self.jobscript.remote_dir} &&' \
              f' {parent.submitter} {self.jobscript.name}'

        if not dry_run:
            parent.transport.transfer()
            parent.url.cmd(cmd, asynchronous=False)
            self.state = Runner._submit_status['final']
        else:
            parent.transport.wipe_transfers()
            self._logger.important(f'launch command: {cmd}')
            self.state = 'dry run'

    def _assess_run(self, **kwargs) -> bool:

        self._extra_files["send"] += kwargs.pop('extra_files_send', [])
        self._extra_files["recv"] += kwargs.pop('extra_files_recv', [])

        self._run_options.update(kwargs)

        self._logger.important(f'assessing run for runner {self}', end='... ')
        self._logger.info('\nrun args:')
        self._logger.info(format_iterable(self.run_args))
        self._logger.info(f'current state is {self.state}')

        if self.run_option('force', False):
            self._logger.warning('force running')
            return True

        if self.state == Runner._submit_status['final']:
            self._logger.warning('skipping already submitted run')
            return False

        # old skip test unpacks a runner from the database
        if self.is_finished and self.run_option('skip', True):
            self._logger.warning('skipping already completed run')
            return False

        self._logger.important('checks passed, running')
        return True

    def _write_runfile(self,
                       dataset,
                       write_file: bool = True) -> str:
        """
        Writes the python file which actually runs the function

        Args:
            dataset:
                parent dataset
            write_file (bool):
                skip writing runfile if false, used only for debugging purposes

        Returns:
            (str): runfile contents
        """

        self._logger.info(f'pre-running function {self.uuid}')

        if self._manual_run:
            self._logger.info('forcing dataset repo write')
            dataset._write_to_repo()

        # check if we have replaced args with a file, and use that if so
        if self._args_replaced:
            argstore = list(self.args.keys())[0]
            argline = f'kwargs = repo.load("' \
                      f'{self._relative_remote_dir(argstore)}")'
        else:
            argline = f'kwargs = {self.args}'

        runscript = [f"repo = __import__('{dataset.repofile}')",
                     argline,
                     f"result = repo.{dataset.function.name}(**kwargs)",
                     f"repo.dump(result, '{self.resultfile.name}')"]

        if dataset.is_child:
            runscript.insert(2, self._dependency_info["parent_import"])
        if dataset.is_parent:
            for line in self._dependency_info['child_submit']:
                runscript.append(line)

        output = '\n'.join(runscript)
        if write_file:
            if not os.path.isdir(self.local_dir):
                self._logger.info(f'creating local dir {self.local_dir}')
                os.makedirs(self.local_dir)
            with open(self.runfile.local, 'w+') as o:
                o.write(output)
            self.state = 'runfile written'
        else:
            self.state = 'runfile write skipped'

        return output

    def _write_script(self,
                      python: str,
                      script,
                      write_file: bool = True) -> str:
        """
        Writes the jobscript for this runner

        Args:
            python (str):
                python command to launch runfile
            script (str, URL.script):
                script header
        """
        tmp = []
        if isinstance(script, str):
            tmp.append(script)
        else:
            tmp.append(script(**self.run_args))

        if self.run_dir and self.run_dir != self.remote_dir:
            cmd = f'pydir=$PWD\n' \
                  f'mkdir -p {self.run_dir} && ' \
                  f'cd {self.run_dir} && ' \
                  f'{python} ${{pydir}}/{self.runfile.name}'
        else:
            cmd = f'{python} {self.runfile.name}'  # run file

        tmp.append(cmd)

        path = self.jobscript.local
        output = '\n'.join(tmp)
        if write_file:
            self._logger.info(f'writing run script to {path}')
            with open(path, 'w+') as o:
                o.write(output)
            self.state = 'jobscript written'
        else:
            self.state = 'jobscript write skipped'

        return output

    @property
    def is_finished(self) -> bool:
        """
        Attempts to determine if this runner has completed its run

        Returns (bool):
            completion status
        """
        fin = hasattr(self, '_result')
        self._logger.info(f'checking finished state of runner '
                          f'{self.short_uuid} -> {fin}')
        return fin

    def update_run_options(self,
                           run_args: dict) -> None:
        """
        Update run args with dict `run_args`
        Args:
            run_args (dict):
                new run arguments

        Returns:
            None
        """
        self._logger.info('updating run options with new run args:')
        self._logger.info(format_iterable(run_args))

        self._run_options.update(run_args)

    def run_option(self,
                   option: str,
                   default=None):
        """
        Return a run option

        Args:
            option (str):
                key to search for
            default:
                default argument to provide to get

        Returns:
            option if available, else None
        """
        ret = self._run_options.get(option, default)
        self._logger.debug(f'getting run option {option}: {ret}')
        return ret

    @property
    def run_args(self) -> dict:
        """
        Display the run arguments

        Returns:
            (dict) run_args
        """
        return self._run_options



"""
URL base class for connecting to remote systems
"""
import logging
import os
import time
import typing

from remotemanager.connection.testing_object import ConnectionTest
from remotemanager.logging.utils import format_iterable
from remotemanager.connection.cmd import CMD
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils import ensure_list
from remotemanager.logging import LoggingMixin
from remotemanager.logging.verbosity import Verbosity


class URL(SendableMixin, LoggingMixin):
    """
    Container to store the url info for a Remote run

    The url should contain everything pertaining to the _remote_, allowing
    Dataset to be remote-agnostic

    Arguments:
        user (str):
            username for the remote system
        host (str):
            host address of the remote system
        port (int, str):
            port to connect to for ssh tunnels
        verbose (bool):
            base-verbosity for connections
        timeout (int):
            time to wait before issuing a timeout for cmd calls
        max_timeouts (int):
            number of times to attempt cmd communication in case of a timeout
        python (str):
            string used to initiate a python instance
        raise_errors (bool):
            set false to ignore errors by default in cmd calls
        passfile (str):
            absolute path to password file for sshpass calls
        envpass (str):
            environment variable containing absolute path to password file for
            sshpass calls
        kwargs:
            any extra args that may end up here from a Dataset or Computer are
            discarded
    """

    _localhost = 'localhost'

    _do_not_package = ['_urlutils']

    def __init__(self,
                 user: str = None,
                 host: str = None,
                 port: int = None,
                 verbose: [int, Verbosity] = None,
                 timeout: int = 5,
                 max_timeouts: int = 3,
                 python: str = 'python',
                 raise_errors: bool = True,
                 keyfile: str = None,
                 passfile: str = None,
                 envpass: str = None,
                 **kwargs):

        self._verbose = Verbosity(verbose)

        if host is None:
            host = URL._localhost
        self._conn = {'user': user,
                      'host': host,
                      'port': port}

        self.timeout = timeout
        self.max_timeouts = max_timeouts
        self._python = python

        # explicit path takes precedent over environment variable
        if passfile is None and envpass is not None:
            passfile = os.environ[envpass]

        self._keyfile = keyfile
        self._passfile = passfile
        self._raise_errors = raise_errors
        self._ssh_override = None

        self._logger.info('new url created with url details:'
                          f'{format_iterable(self._conn)}')

        self._callcount = 0

        self._latency = 0
        self._connection_test = None

    @property
    def raise_errors(self):
        return self._raise_errors

    @raise_errors.setter
    def raise_errors(self, r):
        self._raise_errors = r

    @property
    def verbose(self):
        return self._verbose

    @property
    def call_count(self) -> int:
        return self._callcount

    @property
    def user(self) -> str:
        """
        Currently configured username
        """
        return self._conn['user']

    @user.setter
    def user(self, user):
        """
        Set the user attribute
        """
        self._conn['user'] = user

    @property
    def host(self) -> str:
        """
        Currently configured hostname
        """
        return self._conn['host'] or URL._localhost

    @host.setter
    def host(self, host):
        """
        Set the host attribute
        """
        self._conn['host'] = host

    @property
    def userhost(self) -> str:
        """
        `user@host` string if possible, just `host` if user is not present
        """
        if self.user is None:
            return self.host
        else:
            return f'{self.user}@{self.host}'

    @property
    def python(self):
        return self._python

    @python.setter
    def python(self, python):
        self._python = python

    @property
    def port(self) -> int:
        """
        Currently configured port (defaults to 22)
        """
        port = self._conn['port'] or 22
        return port

    @property
    def ssh(self) -> str:
        """
        ssh insert for commands on this connection
        """
        if self.is_local:
            return ''

        if self._ssh_override:
            return self._ssh_override

        ret = []

        if self.passfile is not None:
            ret.append(f'sshpass -f {self.passfile}')

        ret.append(f'ssh -p {self.port}')

        if self._keyfile:
            ret.append(f'-i {self.keyfile}')

        ret.append(self.userhost)

        return ' '.join(ret)

    @ssh.setter
    def ssh(self,
            newssh: str) -> None:
        """
        Allows forced override of the ssh command

        Inserting extra flags into the ssh can be done as follows:

        >>> url = URL()
        >>> print(url.ssh)
        >>> "ssh"
        >>> url.ssh = "LANG=C " + url.ssh
        >>> print(url.ssh)
        >>> "LANG=C ssh"

        Args:
            newssh (str):
                new ssh string to insert

        Returns:
            None
        """
        self._ssh_override = newssh

    def clear_ssh_override(self):
        """
        Wipe any override applied to ssh. Can also be done by setting
        url.ssh = None

        Returns:
            None
        """
        self._ssh_override = None

    @property
    def keyfile(self):
        if self._keyfile is None:
            return
        p = self._keyfile.replace('~', os.environ['HOME'])
        if not os.path.isfile(p):
            raise RuntimeError(f'could not find ssh key file at '
                               f'{self._keyfile}')

        return self._keyfile

    @property
    def passfile(self):
        if self._passfile is None:
            return

        p = self._passfile.replace('~', os.environ['HOME'])
        if not os.path.isfile(p):
            raise RuntimeError(f'could not find password file at '
                               f'{self._passfile}')

        return self._passfile

    @property
    def is_tunnel(self):
        """
        Uses the presence of an assigned port to determine if we are tunnelling
        or not

        Returns (bool):
           True if this is a tunnelled connection
        """
        return self._conn['port'] is not None

    @property
    def is_local(self):
        """
        True if this connection is purely local
        """
        host = self.host
        if host == URL._localhost:
            return True
        elif host.startswith('127.'):
            return True
        return False

    def ping(self,
             n: int = 5,
             timeout: int = 30) -> float:
        """
        Perform and monitor a ping command

        Args:
            n (int): number of pings to aim for
            timeout (int): kill the process and return 0 if
                this period is elapsed

        Returns:
            (float) latency in ms
        """
        self._logger.runtime(f"pinging {self.host}")

        def cleanup(process):
            process.kill()
            try:
                os.remove("tmp_ping")
                self._logger.debug('removed temporary file')
            except FileNotFoundError:
                self._logger.debug('temporary ping file could not be found')

        ping = self.cmd(f"ping {self.host}",
                        local=True,
                        stdout='tmp_ping',
                        asynchronous=True)

        lines = []
        t0 = time.time()
        while len(lines) < n + 1:
            with open('tmp_ping', 'r') as o:
                lines = o.readlines()

            time.sleep(0.5)

            if time.time() - t0 > timeout:
                self._logger.warning(f'ping timed out after {timeout} seconds')
                cleanup(ping)
                return -1

        cleanup(ping)

        times = []
        for line in lines[1:]:
            times.append(self._process_ping_line(line))

        avg = sum(times) / n
        self._logger.runtime(f'ping times: {times} -> {avg}')
        return avg

    def _process_ping_line(self, line: str) -> float:
        """
        Function to parse the contents of a line output from a ping command

        Args:
            line (str):
                ping command line
        Returns:
            (float): ping time in seconds
        """
        timing = line.split('time=')[1].strip()  # get timing
        val, units = timing.split()
        self._logger.debug(f'parsed line {line} -> {val} ({units})')

        if units != 'ms':
            raise ValueError(f"Unknown ping units {units}")

        return float(val) * 1E-3  # convert to seconds

    def cmd(self,
            cmd: str,
            asynchronous: bool = False,
            local: [bool, None] = None,
            stdout: str = None,
            stderr: str = None,
            timeout: int = None,
            max_timeouts: int = None,
            raise_errors: bool = None,
            dry_run: bool = False) -> typing.Union[CMD, str]:
        """
        Creates and executes a command

        Args:
            asynchronous (bool):
                run this command asynchronously
            cmd (str):
                command to execute
            local (bool, None):
                force a local or remote execution. Defaults to None
            stdout (str):
                optional file to redirect stdout to
            stderr (str):
                optional file to redirect stderr to
            timeout (int):
                time to wait before issuing a timeout
            max_timeouts (int):
                number of times to attempt communication in case of a timeout
            raise_errors (bool):
                override for global setting. Raise any stderr if encountered
            dry_run (bool):
                don't exec the command if True, just returns the string

        Returns (CMD):
            returned command instance
        """
        if local is not None and not local:
            self._logger.info('forced remote call; appending ssh to cmd')
            cmd = f"{self.ssh} '{cmd}'"
        elif local is None and not self.is_local:
            self._logger.info('implicit remote call; appending ssh to cmd')
            cmd = f"{self.ssh} '{cmd}'"
        if raise_errors is None:
            raise_errors = self._raise_errors

        timeout = self.timeout \
            if timeout is None else timeout
        max_timeouts = self.max_timeouts \
            if max_timeouts is None else max_timeouts

        if dry_run:
            return cmd.strip()

        thiscmd = CMD(cmd.strip(),
                      asynchronous=asynchronous,
                      stdout=stdout,
                      stderr=stderr,
                      timeout=timeout,
                      max_timeouts=max_timeouts,
                      raise_errors=raise_errors)
        thiscmd.exec()
        if not local:
            self._callcount += 1
        return thiscmd

    @property
    def utils(self):
        """
        Handle for the URLUtils module
        """
        if self.is_missing('_urlutils'):
            self._urlutils = URLUtils(self)
        return self._urlutils

    def test_connection(self) -> None:
        """
        Create a ConnectionTest instance and run the tests

        Returns:
            None
        """
        testing_object = ConnectionTest(self)
        testing_object.exec()

        self._connection_test = testing_object

    @property
    def connection_test(self) -> ConnectionTest:
        """
        Return the connection test object

        Returns:
            ConnectionTest: testing object
        """
        return self._connection_test

    @property
    def connection_data(self) -> dict:
        """
        Returns the results of a previous connection test

        Returns:
            (dict) connection data
        """
        return self._connection_test.data

    @property
    def latency(self) -> [float, None]:
        """
        Attempts to access the latency property of the stored ConnectionTest

        Returns:
            (float): connection latency in seconds, if available. Else None
        """
        if self._connection_test is not None:
            return self._connection_test.latency
        return None


class URLUtils:
    """
    Extra functions to go with the URL class, called via URL.utils

    As it requires a parent `URL` to function, and is instantiated with a
    `URL`, there is little to no purpose to using this class exclusively

    Arguments:
        parent (URL):
            parent class to provide utils to
    """
    def __init__(self, parent: URL):

        self._logger = logging.getLogger(__name__ + '.URLUtils')
        self._logger.info(f'creating a utils extension to parent: {parent}')

        self._parent = parent

    def file_mtime(self,
                   files: list,
                   local: bool = None,
                   python: bool = False,
                   dry_run: bool = False) -> dict:
        """
        Check file modification times of [files]

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search
            python (bool):
                ensure python style search is used
            dry_run (bool):
                print command only

        Returns (dict):
            {file: mtime (unix)} dictionary
        """

        self._logger.info(f'performing stat on files: {files}')
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)
        times, error = self._file_mtime(files, local, python, dry_run)

        if dry_run:
            # in this instance "times" is simply the command
            return times

        self._logger.info('received:')
        self._logger.info(times)
        self._logger.info(error)
        output = {}
        for file in files:
            if file in times:
                output[file] = times[file]

            else:
                output[file] = None

        return output

    def _file_mtime(self,
                    files: list,
                    local: bool,
                    python: bool,
                    dry_run: bool):
        """
        Perform the "stat -c %Y" command on a list of files,
        returning the result. Uses a python command backup if this fails

        Args:
            files (list):
                list of files to check
            local (bool):
                force a local search
            python (bool):
                force the python override
            dry_run (bool):
                print command only

        Returns:
            (list): list of file unix times
        """
        sep = ','

        def stat():
            self._logger.debug('attempting raw stat command on files')
            basecmd = f'stat -c %n{sep}%Y'
            if len(files) == 1:
                cmd = f'{basecmd} {files[0]}'
            else:
                cmd = f'{basecmd} {{' + ','.join(files) + '}'

            ret = self._parent.cmd(cmd,
                                   local=local,
                                   raise_errors=False,
                                   dry_run=dry_run)

            if dry_run:
                return ret, '', '', ''

            times = {}
            for line in ret.stdout.split('\n'):
                try:
                    times[line.split(sep)[0]] = int(float(line.split(sep)[1]))
                except IndexError:
                    pass

            return times, ret.stderr.split('\n'), ret.returncode, ret.stderr

        def pystat():
            self._logger.debug('attempting python stat on files')
            ex = f"""import os
files={files}
for f in files:
\ttry: print(f'{{f}}{sep}{{os.stat(f).st_mtime}}') 
\texcept FileNotFoundError: print(f)"""

            cmd = f'{self._parent.python} -c "{ex}"'

            ret = self._parent.cmd(cmd,
                                   local=local,
                                   raise_errors=False,
                                   dry_run=dry_run)

            if dry_run:
                return ret, None

            times = {}
            error = []
            for line in ret.stdout.split('\n'):
                try:
                    times[line.split(sep)[0]] = int(float(line.split(sep)[1]))
                except IndexError:
                    error.append(line.strip())

            return times, error

        files = ensure_list(files)

        if not python:
            t, e, returncode, stderr = stat()
            if returncode in [126, 127] or 'illegal option' in stderr:
                self._logger.warning('stat failed, falling back on python')
                return pystat()

            return t, e

        return pystat()

    def file_presence(self,
                      files: list,
                      local: bool = None,
                      dry_run: bool = False) -> dict:
        """
        Search for a list of files, returning a boolean presence dict

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search
            dry_run (bool):
                print command only

        Returns (dict):
            {file: present} dictionary
        """
        self._logger.info(f'checking for presence of files: {files}')
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)

        times = self.file_mtime(files,
                                local=local,
                                dry_run=dry_run)

        if dry_run:
            return times

        return {f: times[f] is not None for f in files}

    def search_folder(self,
                      files: list,
                      folder: str,
                      local: bool = None,
                      dry_run: bool = False) -> dict:
        """
        Search `folder` for `files`, returning a boolean presence dict

        Arguments:
            files (list):
                list of filenames to check for. Optionally, a string for a
                single file
            folder (str):
                folder to scan
            local (bool):
                perform the scan locally (or remotely)
            dry_run (bool):
                print command only

        Returns (dict):
            {file: present} dictionary
        """
        if local is None:
            local = self._parent.is_local
        fpath = os.path.abspath(folder) if local else folder

        self._logger.debug(f'scanning folder {fpath}')
        self._logger.debug('searching for files:')
        self._logger.debug(f'{format_iterable(files)}')

        ls_return = self.ls(fpath,
                            local=local,
                            as_list=True,
                            dry_run=dry_run)

        if dry_run:
            self._logger.info('dry run, returning command')
            return ls_return

        scan = [os.path.basename(f) for f in ls_return]

        self._logger.debug('scan sees:')
        self._logger.debug(f'{format_iterable(scan)}')

        if isinstance(files, str):
            self._logger.info('files is a string, running in singular mode')
            ret = {files: os.path.basename(files) in scan}
        else:
            ret = {file: os.path.basename(file) in scan for file in files}

        return ret

    def touch(self,
              file: str,
              local: bool = None,
              raise_errors: bool = None,
              dry_run: bool = False) -> CMD:
        """
        perform unix `touch`, creating or updating `file`

        Arguments:
            file (str):
                filename or path to file
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f'utils touch on file {file}')
        fname = os.path.abspath(file) if local else file
        return self._parent.cmd(f'touch {fname}',
                                local=local,
                                raise_errors=raise_errors,
                                dry_run=dry_run)

    def mkdir(self,
              file: str,
              local: bool = None,
              raise_errors: bool = None,
              dry_run: bool = False) -> CMD:
        """
        perform unix `mkdir -p`, creating a folder structure

        Arguments:
            file (str):
                name or path to folder
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f'utils mkdir on path {file}')
        fname = os.path.abspath(file) if local else file
        # print(f'making dir {fname}')
        return self._parent.cmd(f'mkdir -p {fname}',
                                local=local,
                                raise_errors=raise_errors,
                                dry_run=dry_run)

    def ls(self,
           file: str,
           as_list: bool = True,
           local: bool = None,
           raise_errors: bool = None,
           dry_run: bool = False) -> [CMD, list]:
        """
        Identify the files present on the directory

        Arguments:
            file (str):
                name or path to folder.
            as_list (bool):
                convert to a list format
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD, list):
            CMD instance for the command, or the list if as_list is True
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f'utils ls on path {file}')
        fname = os.path.abspath(file) if local else file

        ret = self._parent.cmd(f'ls {fname}',
                               local=local,
                               raise_errors=raise_errors,
                               dry_run=dry_run)

        if as_list and not dry_run:
            ret = [f for f in ret.stdout.split('\n') if f != '']
        return ret



INDENT_CHAR = '   '


def format_iterable(iterable,
                    exclude=None,
                    indent: int = 1) -> str:
    """
    Recursive formatting of an iterable, producing log-safe string

    Arguments:
        iterable (list, tuple, set, dict):
            iterable to sort
        exclude (list, optional):
            keys to exclude
        indent (int, optional):
            indent level for recursion

    Returns:
        (str):
            newline separated string
    """
    def treat_dict(inp):
        ret = []
        for a in sorted(inp.keys()):
            v = inp[a]

            if a not in exclude:
                if type(v) in dispatch.keys():
                    v = format_iterable(iterable=v,
                                        exclude=exclude,
                                        indent=indent + 1)
                ret.append(INDENT_CHAR * indent + f'{a}: {v}')

        return '\n' + '\n'.join(ret)

    def treat_iterable(inp):
        exclude.append('')

        ret = []
        for v in inp:

            if v not in exclude:
                if type(v) in dispatch.keys():
                    v = format_iterable(iterable=v,
                                        exclude=exclude,
                                        indent=indent + 1)
                ret.append(INDENT_CHAR * (indent-1) + f'- {v}')

        return f'\n' + f'\n'.join(ret)

    def treat_printable(inp):
        if inp is None:
            return 'None'
        return f'{inp}'

    # can hit empty lists (often in recursion)
    # output looks nicer if they're output without formatting
    try:
        if len(iterable) == 0:
            return '\n' + INDENT_CHAR * indent + f'{iterable}'
    except TypeError:
        return treat_printable(iterable)

    if exclude is None:
        exclude = []
    elif isinstance(exclude, str):
        exclude = [exclude]

    dispatch = {dict: treat_dict,
                list: treat_iterable,
                tuple: treat_iterable,
                set: treat_iterable,
                str: treat_printable,
                type: treat_printable,
                }
    return dispatch[type(iterable)](iterable)



from .dataset.dataset import *  # noqa: F403
from .connection.url import *  # noqa: F403
from .logging.log import *  # noqa: F403
from .storage.remotefunction import *  # noqa: F403

__all__ = ['Dataset', 'URL', 'RemoteFunction']   # noqa: F405
__version__ = '0.5.12'

# attach a global Logger to the manager
Logger = Handler()    # noqa: F405


# ipython magic
def load_ipython_extension(ipython):
    from remotemanager.jupyter.magic import RCell
    ipython.register_magics(RCell)



import inspect

from remotemanager import URL
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.logging import LoggingMixin


class BaseComputer(URL):
    """
    Base computer module for HPC connection management.

    Extend this class for connecting to your machine

    Adds `module_purge`, set to False to ignore the pre-module
    `module purge` call.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.shebang = '#!/bin/bash'
        self.pragma = NotImplemented

        self.module_purge = kwargs.get('module_purge', True)

        self.modules = kwargs.get('modules', [])

        self._extra = ''
        self._internal_extra = ''

    def __setattr__(self, key, value):
        """
        If the set `key` attribute is an MPI option, instead set the `value`
        of that attribute

        Args:
            key:
                attribute name to set
            value:
                value to set to

        Returns:
            None
        """
        if key in self.__dict__:
            if isinstance(getattr(self, key), (optional, required)):
                getattr(self, key).value = value
                return

        object.__setattr__(self, key, value)

    @property
    def arguments(self):
        return [k for k, v in self.__dict__.items()
                if isinstance(v, (optional, required))]

    @property
    def argument_dict(self):
        return {k.strip(): getattr(self, k) for k in self.arguments}

    @property
    def required(self):
        """
        Returns the required arguments
        """

        # grab the required arguments for the class, and fill with placeholders
        # of the correct type
        params = inspect.signature(self.__class__).parameters
        required_args = {}
        for arg_name, v in params.items():
            if v.default is inspect._empty and arg_name != 'kwargs':
                required_args[arg_name] = v.annotation

        if 'dummy' in params.keys():
            required_args['dummy'] = True

        temp = self.__class__(**required_args)
        return [k for k, v in temp.__dict__.items() if isinstance(v, required)]

    @property
    def missing(self):
        """
        Returns the currently missing arguments
        """
        return [k for k in self.required if not getattr(self, k)]

    @property
    def valid(self):
        return len(self.missing) == 0

    def update_resources(self, **kwargs):
        """
        Set any arguments passed to the script call

        Args:
            **kwargs:
                kwarg updates e.g. mpi=128
        Returns:
            None
        """
        for k, v in kwargs.items():
            setattr(self, k, v)

    def resources_block(self, **kwargs):

        if 'name' in kwargs:
            # Dataset `name` param detected, use as a default
            if not hasattr(self, 'jobname') or 'jobname' not in kwargs:
                kwargs['jobname'] = kwargs.pop('name')

        self.update_resources(**kwargs)

        if not self.valid:
            raise RuntimeError(f'missing required arguments: {self.missing}')

        options = {}
        for k, v in self.argument_dict.items():
            if v:
                options[v.flag] = v.value

        return [f'{self.pragma} {k}={v}'
                for k, v in sorted(options.items()) if v]

    def modules_block(self):
        base = []
        if self.module_purge:
            base.append('\nmodule purge')
        return base + [f'module load {m}' for m in self.modules]

    @property
    def extra(self):
        return self._internal_extra + self._extra

    @extra.setter
    def extra(self, external):
        self._extra = external

    def script(self,
               **kwargs) -> str:
        """
        Takes job arguments and produces a valid jobscript

        Returns:
            (str):
                script
        """
        script = [self.shebang]

        script += self.resources_block(**kwargs)
        script += self.modules_block()

        if hasattr(self, 'postscript') and self.postscript is not None:
            script.append(self.postscript)

        if hasattr(self, 'extra') and self.extra is not None:
            script.append(self.extra)

        return '\n'.join(script)


class placeholder_option(SendableMixin, LoggingMixin):
    """
    .. warning::
        This class is intended to be subclassed by the optional and required
        placeholders.

    Stub class to sit in place of an option within a computer.

    Args:
        mode (string):
            storage mode, required or optional
        flag (string):
            flag to append value to (`--nodes`, `--walltime`, etc)
    """

    def __init__(self, mode, flag, min, max):
        self._mode = mode
        self._flag = flag
        self._value = None
        self._bool = False

        self._min = min
        self._max = max

    def __hash__(self):
        return hash(self._mode)

    def __str__(self):
        return str(self.value)

    def __bool__(self):
        """
        Makes objects "falsy" if no value has been set, "truthy" otherwise
        """
        return self.value is not None

    @property
    def flag(self):
        return self._flag

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def value(self):
        if hasattr(self, 'default') and self._value is None:
            return self.default

        return self._value

    @value.setter
    def value(self, value):
        try:
            value / 1
            isnumeric = True
        except TypeError:
            isnumeric = False

        if isnumeric:
            if self.min is not None and value < self.min:
                raise ValueError(f'{value} for {self.flag} '
                                 f'is less than minimum value {self.min}')
            if self.max is not None and value > self.max:
                raise ValueError(f'{value} for {self.flag} '
                                 f'is more than maximum value {self.max}')

        self._bool = True
        self._value = value


class required(placeholder_option):
    """
    .. warning::
        This class is intended to be subclassed by the optional and required
        placeholders.

    Stub class to sit in place of an option within a computer.

    This option is _required_, and should raise an error if no value is found

    Args:
        flag (string):
            flag to append value to (`--nodes`, `--walltime`, etc)
    """

    def __init__(self, flag, min=None, max=None):
        super().__init__('required', flag, min, max)


class optional(placeholder_option):
    """
    .. warning::
        This class is intended to be subclassed by the optional and required
        placeholders.

    Stub class to sit in place of an option within a computer.

    This option is not required, and should have an accessible default if
    no value is found

    Args:
        flag (string):
            flag to append value to (`--nodes`, `--walltime`, etc)
        default:
            default value to use if none is assigned
    """

    def __init__(self, flag, default=None, min=None, max=None):
        super().__init__('optional', flag, min, max)

        self._default = default

    @property
    def default(self):
        try:
            return self._default()
        except TypeError as E:
            self._logger.warning(f'recieved error from default attempt: {E}')
            return self._default



"""
Handles file transfer via the `rsync` protocol
"""
import pathlib

from remotemanager.transport.transport import Transport


class rsync(Transport):
    """
    Class for `rsync` protocol

    Args:
        checksum (bool):
            Adds checksum arg, which if True will add --checksum flag to
            parameters
    """
    def __init__(self,
                 *args,
                 **kwargs):

        super().__init__(*args, **kwargs)

        # flags can be exposed, to utilise their flexibility
        flags = kwargs.pop('flags', 'auv')
        self.flags = flags

        if kwargs.get('checksum', False):
            print('adding checksum to rsync')
            self.flags += '--checksum'

        self._logger.info('created new rsync transport')

        self._cmd = 'rsync {flags} {password}{inner_dir}{primary} {secondary}'

    def cmd(self, primary, secondary):

        if self.url.passfile and self.url.keyfile:
            raise RuntimeError('rsync appears to have an issue when '
                               'specifying sshpass AND ssh-key. Either set up '
                               'your ssh config and remove the keyfile or use '
                               'transport.scp')

        def get_remote_dir(path):
            if ':' not in path:
                return path
            return path.split(':')[1]

        password = ''
        if self.url.passfile is not None:
            password = f'--rsh="sshpass -f {self.url.passfile} ssh" '

        cmd = self._cmd
        inner_dir = ''
        if len(pathlib.Path(secondary).parts) > 1:
            # the target is a nested dir. If the whole tree doesn't exist,
            # rsync with throw an error
            if ':' in secondary:
                # target is a remote folder, use the --rsync-path hack
                inner_dir = f'--rsync-path="mkdir -p ' \
                            f'{get_remote_dir(secondary)} && rsync" '
            else:
                cmd = f'mkdir -p {secondary} && {self._cmd}'

        base = cmd.format(flags=self.flags,
                          password=password,
                          primary=primary,
                          secondary=secondary,
                          inner_dir=inner_dir)
        self._logger.debug(f'returning formatted cmd: "{base}"')
        return base
